/*
	Front-end to compile .tlcc program.
	Works like gcc so one can do

	cctlcc toto.tlcc -o toto
or
	cctlcc -c toto.tlcc -o toto.o

	.tlcc are converted to .cpp files and then deleted
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/wait.h>
#include <sys/stat.h>

static void check_exist (const char *path)
{
	struct stat st;
	if (lstat(path,&st)!=-1){
		fprintf (stderr,"file %s exist, won't overwrite, aborting\n",path);
		exit (-1);
	}
}

static int cctlcc_exec (
	const char *cmd,
	const char *arg1,
	const char *arg2,
	char *opts[])
{
	int ret = 0;
	pid_t pid = fork();
	if (pid == 0){
		int nb=0;
		if (opts != NULL) for (; opts[nb] != NULL; nb++);
		const char *argv[nb+4];
		argv[0] = cmd;
		for (int i=0; i<nb; i++) argv[i+1] = opts[i];
		argv[nb+1] = arg1;
		argv[nb+2] = arg2;
		argv[nb+3] = NULL;
		// for (int i=0; i<nb+3; i++) fprintf (stderr,"argv[%d]=%s\n",i,argv[i]);
		execvp (argv[0],(char* const*)argv);
	}else if (pid == (pid_t)-1){
		ret = -1;
		fprintf (stderr,"Can't fork (%s)\n",strerror(errno));
	}else{
		int status;
		while (wait(&status)!=-1);
		if (WIFEXITED(status)) ret = WEXITSTATUS(status);
	}
	return ret;
}


int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc <= 1){
		fprintf (stderr
			,"cctlcc: tlmp gcc wrapper\n"
			 "\n"
			 "\tIt translates .tlcc files into .cpp C++ files\n"
			 "\tand calls the C++ compiler to complete the tasks.\n"
			 "\tAll option are passed to the C++ compiler. cctlcc has\n"
			 "\tno option of its own.\n"
			 "\n"
			 "\tBy default \"gcc\" is used as the C++ compiler.\n"
			 "\tThis may be overriden by setting the environment variable\n"
			 "\tCC: CC=g++ cctlcc ... foo.tlcc ...\n"
			 "\n"
			);
	}else{
		char *newargv[argc+1];
		char *tmpfiles[2*argc+1];
		int nbtmp = 0;
		int lenopt = 0;
		for (int i=1; i<argc; i++) lenopt += strlen(argv[i])+1;
		char cppoptions[lenopt+1];
		cppoptions[0] = '\0';
		bool tlmpoptions = true;
		bool c_opt_seen = false;
		int dst = 1;
		ret = 0;
		for (int i=1; i<argc && ret == 0; i++){
			char *arg = argv[i];
			if (strcmp(arg,"--notlmpoptions")==0){
				tlmpoptions = false;
			}else if (strncmp(arg,"-I",2)==0
				|| strncmp(arg,"-D",2)==0){
				strcat (cppoptions," ");
				strcat (cppoptions,arg);
				newargv[dst++] = strdup(arg);
			}else if (strcmp(arg,"-c")==0){
				newargv[dst++] = strdup(arg);
				c_opt_seen = true;
			}else{
				const char *end = strstr(arg,".tlcc");
				if (end != NULL && end[5] == '\0'){
					// We must convert this file in .cc using the tlcc utility
					char path_version[PATH_MAX],path_cpp[PATH_MAX];
					strncpy (path_version,arg,sizeof(path_version)-1);
					strncpy (path_cpp,arg,sizeof(path_cpp)-1);
					strcpy (path_cpp+(end-arg),".cpp");
					strcpy (path_version+(end-arg),".version");
					char tmp_version[PATH_MAX];
					const char *sl = strrchr(path_version,'/');
					if (sl == NULL){
						sl = path_version;
					}else{
						sl++;
					}
					snprintf (tmp_version,sizeof(tmp_version),"/tmp/%s.%d",sl,getpid());
					check_exist (tmp_version);
					char tmpoptions[strlen(cppoptions)+200];
					if (tlmpoptions){
						snprintf (tmpoptions,sizeof(tmpoptions)-1
							,"-DEXPORT= -I/usr/include/tlmp -I/usr/include/linuxconf %s"
							,cppoptions);
					}else{
						strcpy (tmpoptions,cppoptions);
					}
					char *cppopts[]={"--cppoptions",tmpoptions,NULL};
					cctlcc_exec ("tlccversion",arg,tmp_version,cppopts);
					tmpfiles[nbtmp++] = strdup(tmp_version);
					char *tlccopts[]={
						"--compvers",tmp_version,
						"--compvers","TLMP.version",
						"--compvers",path_version,
						NULL
					};
					check_exist (path_cpp);
					ret = cctlcc_exec ("tlcc",arg,path_cpp,tlccopts);
					newargv[dst++] = strdup(path_cpp);
					tmpfiles[nbtmp++] = strdup(path_cpp);
				}else{
					newargv[dst++] = strdup(arg);
				}
			}
		}
		newargv[dst] = NULL;
		{
			char *gcc = (char*)getenv ("CC");
			newargv[0] = gcc == NULL ? (char*)"gcc" : gcc;
		}
		if (ret == 0){
			pid_t pid = fork();
			if (pid == 0){
				if (tlmpoptions){
					char *moreargv[argc+1+4];
					moreargv[0] = newargv[0];
					moreargv[1] = "-DEXPORT=";
					moreargv[2] = "-I/usr/include/tlmp";
					moreargv[3] = "-I/usr/include/linuxconf";
					for (int i=1; i<=argc; i++){
						moreargv[i+3] = newargv[i];
					}
					if (!c_opt_seen){
						moreargv[argc+3] = "-llinuxconf";
						moreargv[argc+4] = NULL;
					}
					execvp (moreargv[0],moreargv);
				}else{
					execvp (newargv[0],newargv);
				}
				fprintf (stderr,"Can't execute gcc (%s)\n",strerror(errno));
				_exit (-1);
			}else if (pid == (pid_t)-1){
				fprintf (stderr,"Can't fork (%s)\n",strerror(errno));
			}else{
				int status;
				while (wait(&status)!=-1);
				if (WIFEXITED(status)){
					ret = WEXITSTATUS(status);
				}
			}
		}
		for (int i=0; i<nbtmp; i++){
			char *tmpfile = tmpfiles[i];
			//fprintf (stderr,"unlink %s\n",tmpfile);
			unlink (tmpfile);
			free (tmpfile);
		}
	}
	return ret;
}

